;;; File: causal-order-aux.lisp
;;; Author: Taoxin Peng & Norbert Eschle
;;; Last Revised: August 5 1999

;; ------------------------------------------------------------------------
;; The functoins down here are the predicates used for the causal ordering

;; CD. Degree of commitment. It calculates the degree of commitment for a
;; variable. This is being done by 'counting all the dimensions of the
;; variable which are not equal to zero. If we have negative time then don't
;; take it into account.
;; The functino takes one variable frame as input.
(defun cd (var)
  
  ;; Get the dimension frame of the variable. Also check for negative time and
  ;; prepare the output-value.
  (let* ((dimension-frame (car (gv var :has-dimensions)))
	 (neg-time-flag (> 0 (gv dimension-frame :t)))
	 (output-value 0))
    
    ;; If negative time then ...
    (if neg-time-flag
	
	;; recurse over the variables without the time-dimension slot.
	(doslots (dimension dimension-frame)
		 (if (and (numberp (gv dimension-frame dimension))
			  (not (= 0 (gv dimension-frame dimension)))
			  (not (eq dimension :t)))
		     (setf output-value (1+ output-value))))
      
      ;; ... if not then count it in as a normal dimension.
      (doslots (dimension dimension-frame)
	       (if (and (numberp (gv dimension-frame dimension))
			(not (= 0 (gv dimension-frame dimension))))
		   (setf output-value (1+ output-value)))))
    
    ;; Return the output value
    output-value))


;; Degree of factorisation excluding time.
(defun fd-ext (var)
  
  ;; Prepare the output-value and get the dimension frame of the variable.
  (let* ((output-value 0)
	 (dimension-frame (car (gv var :has-dimensions))))
    
    ;; recurse over the slots of the dimension frame and do not consider
    ;; the :t slot but consider every other slot which has a number in it
    ;; since they are dimensions of the variable.
    (doslots (dimension dimension-frame)
	     (if (and (numberp (gv dimension-frame dimension))
		      (not (eq dimension :t)))
		 
		 ;; If there is a dimension then add its power to the output-
		 ;; value.
		 (setf output-value (+ output-value
				       (gv dimension-frame dimension)))))
    
    ;; Return the output-value.
    output-value))


;; Degree of factorisation. This is the same as fd-ext but also considers time.
(defun fd (var)
  (let* ((output-value 0)
	 (dimension-frame (car (gv var :has-dimensions))))
    (doslots (dimension dimension-frame)
	     (if (numberp (gv dimension-frame dimension))
		 (setf output-value (+ output-value
				       (gv dimension-frame dimension)))))
    
    ;; Return the output-value
    output-value))


;; Beless. This predicate decides whether its first argument - a variable - is
;; less independent than its second argument, another variable.
(defun beless (var-1 var-2)
  
  ;; Return t if either
  (if
      
      ;; the cd of var 1 is greater than that of var 2.
      (or (> (gv var-1 :cd) (gv var-2 :cd))
	  
	  ;; or both cds are equal and the fd-ext of var 1 is greater
	  (and (= (gv var-1 :cd) (gv var-2 :cd))
	       (> (gv var-1 :fd-ext) (gv var-2 :fd-ext)))
	  
	  ;; or both cd and fd-ext are equal and the fd of var 1 is greater.
	  (and (= (gv var-1 :cd) (gv var-2 :cd))
	       (= (gv var-1 :fd-ext) (gv var-2 :fd-ext))
	       (> (gv var-1 :fd) (gv var-2 :fd))))
      t
    
    ;; else nil.
    nil))


;; This function sets all the values of fd, cd and fd-ext to 0. It is called
;; by the actual ordering algorithm.
(defun set-zero (var)
  (s-value var :cd 0)
  (s-value var :fd 0)
  (s-value var :fd-ext 0))


;; This function here prepares the variables for processing. It sets the 
;; necessary values (cd, fd, fd_ext) to perform causal ordering. It does not
;; take any argument and outputs only t.
(defun prepare-vars ()
  
  ;; First get the variables.
  (let ((var-list (gv variable :is-a-inv)))
    
    ;; Now mapc for every variable
    (mapc #'(lambda (var)
	      (s-value var :cd (cd var))
	      (s-value var :fd (fd var))
	      (s-value var :fd-ext (fd-ext var)))
	  var-list))
  t)


;; This function creates the derivative of a variable. It creates a schema for
;; the drivative variable and its dimensions using the normal kb-types for this
;; but links these schemas to the base with the :derivative-of and 
;; :derivative-dimensions-of relations. These relations trigger inheritance.
;; The function takes one argument, the variable.
(defun build-derivative (var)
  (let* (
	 
	 ;; Get the dimensions of the variable.
	 (dimensions-of-var (car (gv variable :has-dimensions)))
	 
	 ;; Build the schema for the derivative variable.
	 (derivative-schema (create-instance (gensym) variable
					     
					     ;; The name of it is
					     ;; "d*<variable-name>"
					     (:name (concatenate 'string
						      "d*"
						      (gv var :name)))
					     
					     ;; Let the relation point to the
					     ;; base.
					     (:derivative-of (list variable))))
	 
	 ;; Now build the subframe for the dimensions of the derivative.
	 (derivative-dimensions (create-instance (gensym) dimensions
						 
						 ;; Let it point towards the
						 ;; derivative schema which it
						 ;; is the subframe of.
						 (:dimensions-of
						  (list derivative-schema))
						 
						 ;; Point towards the
						 ;; dimension subframe of the
						 ;; base to inherit the
						 ;; values.
						 (:derivative-dimensions-of
						  (list dimensions-of-var))
						 
						 ;; manipulate the time dim:
						 ;; If it is not present then
						 ;; set it to -1. Else it is
						 ;; the base value -1.
						 (:t (if (gv dimensions-of-var
							     :t)
							 (1- (gv 
							      dimensions-of-var
							      :t))
						       (- 1))))))
    
    ;; Return the derivative.
    derivative-schema))


;; This function takes a list of variables as its input. It then returns a list
;; of the derivatives of the variables in the input list.
(defun make-list-of-derivatives (list-of-vars)
  (remove-if #'null
	     (mapcar #'(lambda (var)
			 (if (derivative-p var)
			     var))
		     list-of-vars)))


;; This function gets a list of variables and only returns those that are
;; not derivatives.
(defun make-list-of-non-derivatives (list-of-vars)
  (remove-if #'null
	     (mapcar #'(lambda (var)
			 (if (not (derivative-p var))
			     var))
		     list-of-vars)))


;; This function tells us whether a variable is a derivative or not.
(defun derivative-p (var)
  (if (and (eq (elt (gv var :name) 0)
	       #\d)
	   (eq (elt (gv var :name) 1)
	       #\*))
      t))


;; This function partitions a list of variables into variables and derivatives
;; present in the input list. It has two return values: the first is the list
;; of variables and the second is the list of derivatives.
(defun partition-list (list-of-vars)
  (let ((list-vars (make-list-of-non-derivatives list-of-vars))
	(list-derivs (make-list-of-derivatives list-of-vars)))
    (values list-vars list-derivs)))


;; This function takes a list of variables and returns two lists where the
;; first list is the left hand side of an equation and the second list is the
;; RHS of an equation. The most dependent variables in the input list will be
;; in the first list. This is determined by the use of the predicate beless.
;; The first return value is the lhs of a symbolic equation and the second is
;; the rhs of an equation.
(defun make-equation (list-of-vars)
  (multiple-value-bind (list-vars list-derivs)
      (partition-list list-of-vars)
    
    ;; Sort the lists of derivatives and variables using the predicate beless.
    (let* ((sorted-list-vars (sort list-vars #'beless))
	   (sorted-list-derivs (sort list-derivs #'beless)))
      
      ;; if derivatives are present then partition the list into lhs and rhs
      ;; which consists of the list of derivatives together with the variables.
      (if sorted-list-derivs
	  (make-lhs-rhs (append sorted-list-derivs sorted-list-vars))
	
	;; Else only partition the sorted list of variables.
	(make-lhs-rhs sorted-list-vars)))))


;; This function here splits a sorted list into lhs and rhs of a symbolic eq.
;; The lhs-list takes the leftmost members of the input list for which holds:
;; !beless(x,y). This means that they have the same degree of dependency.
(defun make-lhs-rhs (list-of-vars)
  
  ;; Iterate over the input list. The test-var is the most dependent variable.
  ;; It is the car of the input list. This variable does not get updated in the
  ;; whole loop.
  (do* ((test-var (car list-of-vars))
	
	;; The rhs is updated by the following:
	;; If !beless(test-var,1st-var-of-rhs) then remove this var from the
	;; rhs-list. else nothing changes.
	(rhs (cdr list-of-vars)
	     (if rhs
		 (if (not (beless test-var (car rhs)))
		     (cdr rhs)
		   rhs)
	       rhs))
	
	;; The lhs is updated by:
	;; If the above condition for the rhs-update is t then stick the first
	;; variable of the rhs to the end of lhs. Else nothing changes.
	(lhs (list test-var)
	     (if rhs
		 (if (not (beless test-var (car rhs)))
		     (append lhs (list (car rhs)))
		   lhs)
	       lhs)))
      
      ;; Terminate the loop when nothing changes (beless returns t).
      ((or (null rhs) (beless test-var (car rhs))) ;Added the disjunction and 
					; the null to deal with error.
       
       ;; Return BOTH lhs and rhs in that order.
       (if rhs (values lhs rhs)))))	;Added the condition to deal with
					;error. 


;; This function performs the causal ordering of the variables of one symbolic
;; equation. Its input is the list of all variables in one symbolic equation
;; which is input to the whole algorithm.
;; The function establishes symbolic equations between all variables in the
;; symbolic equation. It calls itself recursive until all variables have been
;; processed. Its output is a list of equations. An equation is a list of two
;; elements where every element is a list of variables. The first list is the
;; LHS of the equation and the second the RHS.
(defun create-symbolic-equations (list-of-vars)
  
  ;; If the list of variables is non-nil then
  (if (< 1 (length list-of-vars))
      
      ;; let the equation be two lists returned by make-equation. 
      (let* ((equation (multiple-value-bind (lhs rhs)
			   (make-equation list-of-vars)
			 (list lhs rhs)))
	     
	     ;; The list with which the function will call itself is the cadr
	     ;; of the rhs that is returned by the above funcall.
	     (rhs-list (cadr equation)))
	
	;; Return the list of the equation which has just been computed and
	;; the return value of the recursive call with the rhs-list.
	(cons equation (create-symbolic-equations rhs-list)))
    
    ;; If the argument is nil then return nil.
    nil))


;; This function takes a list of variables as input. It then returns a set of
;; symbolic equations of the form [x] = [d*x]. The lhs is a list and the rhs
;; is a list. The equations is a list of lhs and rhs (ordered). The function 
;; returns a list of equations.
(defun integral-causality (list-of-vars)
  
  ;; Partition the list of variables into derivatives and 'normal' variables.
  (multiple-value-bind (list-vars list-derivs)
      (partition-list list-of-vars)
    
    ;; For every derivative build the equation that it is dependent upon its
    ;; base. Return the list of these equations.
    (mapcar #'(lambda (derivative)
		(list
		 (list (kr-send knowledge-base
				:get-kb-item
				`((:name ,(string-left-trim
					   "d*"
					   (gv derivative :name))))))
		 (list derivative)))
	    list-derivs)))


;; This function performs the causal ordering. It calls the above functions.
;; Input is a list of variables from ONE input equation.
(defun causal-order-eq (list-of-vars)
  (append (create-symbolic-equations list-of-vars)
	  (integral-causality list-of-vars)))
