;;; File: module.lisp
;;; Author: Norbert Eschle
;;; Last revised: 6/10/1999


;;; This file holds all the functions for including the causal ordering
;;; algorithm into Cot 2.0

;; ---------------------------------------------------------------------------
;; Here are the top-level functions.

;; This function wraps the macro call kr-send into a function to avoid problems
;; with the compilation.
(defun retrieve-from-kb (name-string)
  (kr-send knowledge-base :get-kb-item 
	   (list (list :name name-string))))


;; This is the top-level function as specified in the DOcu of Cot 2.0. It
;; starts the causal ordering algorithm and creates the equation file with the
;; causal relations.
(defun start-causal-ordering-algorithm (file)
  
  ;; Use this package for the communication with the operating system.
  (use-package 'excl)
  
  ;; Load the causal ordering functions.
  (load (concatenate 'string 
	  *prg-dir*
	  "/../resources/causal-order-aux.lisp"))
  
  ;; Get the list of input vars to the device, the list of structural
  ;; constraints and the list of symbolic equations. All the equations and the
  ;; constraints contain strings that have to be converted to the corresponding
  ;; objects from the kb. This is done using maptree.
  (let ((input-vars (mapcar #'(lambda (var)
				(retrieve-from-kb
				 var))
			    (read file)))
	(struct-constraints (maptree 
			     #'(lambda (var)
				 (retrieve-from-kb
				  var))
			     (read file)
			     #'mapcar))
	(symbolic-equations (maptree 
			     #'(lambda (var)
				 (retrieve-from-kb
				  var))
			     (read file)
			     #'mapcar)))
    
    ;; First initialise the vars and input vars of the device.
    (prepare-vars)
    (mapc #'set-zero input-vars)
        
    ;; Create the list of causal links as specified in Cot 2.0
    (let ((output-list nil))
      
      ;; Structural constraints.
      (mapc #'(lambda (equation)
		(setf output-list
		  (append output-list (make-output-eq equation))))
	    struct-constraints)
      
         ;; Symbolic equations.
      (setf output-list
	(append (make-causal-links symbolic-equations)
		output-list))
      
      ;; Add the impact sign to the causal links.
      (setf output-list
	(add-impact-sign output-list))
      
      ;; Now store everything to a file.
      (with-open-file (output-file (concatenate 'string
				     *prg-dir*
				     "/../data/output.tmp")
		       :direction :output
		       :if-exists :supersede
		       :if-does-not-exist :create)
	(format output-file "~a"
		
		;; Convert the elements from objects to strings again.
		(maptree #'(lambda (x)
			     (if (schema-p x)
				 (gv x :name)
			       x))
			 output-list
			 #'mapcar)))))
  (let ((command-line (concatenate 'string 
			*prg-dir*
			"/../data/output.tmp")))
    (run-shell-command (concatenate 'string
			 (concatenate 'string
			   *prg-dir*
			   "/../resources/file-proc ")
			 command-line))
    (run-shell-command (concatenate 'string
			 "rm "
			 (concatenate 'string
			 *prg-dir*
			 "/../data/output.tmp"))))
  ;; Clean up the package we loaded
  (unuse-package 'excl))


;; ---------------------------------------------------------------------------
;; Down here are the file functions. They read everything from the file and 
;; pass it to the appropriate back end. They also write everything to the file.


;; ---------------------------------------------------------------------------
;; Here are the functions building the irreversible symbolic equations.

;; This function takes a list representing a symbolic equation as its argument
;; and outputs a list of equations that conform to the specifications of
;; Cot 2.0 eqn-files.
(defun make-output-eq (equation)
  
  ;; First take the LHS and the RHS of the equation.
  (let ((lhs (car equation))
	(rhs (cadr equation))
	
	;; Now prepare an output variable. 
	(output-list nil))
    
    ;; Get all the standard cot 2.0 equations from this function down here.
    ;; This is done by storing the return value of append in the output-list.
      (mapc #'(lambda (x)
		(setf output-list
		  (append output-list (process-lhs-var x rhs))))
	    lhs)    
    ;; We don't want to have any duplicates.
    (remove-duplicates output-list :test #'equal)))


;; This function takes a var from an lhs of an input equation from 
;; make-output-eq and the rhs of this equation and creates as many equations
;; from the possible combinations as possible. The list of irreversible
;; equations is returned.
(defun process-lhs-var (lhs-var rhs)
    (mapcar #'(lambda (x)
		(list (list lhs-var) (list x)))
	    rhs))


;; This is the function which creates the output causal links for the Cot 2.0
;; specifications from the causal equations. Its input is the list of the 
;; symbolic equations modelling device behaviour as specified in the file
;; specs. The output is the list of causal links. 
(defun make-causal-links (equations)
  (do* ((equation-list equations (cdr equation-list))
	(active-equation (car equation-list) (car equation-list))
	(rough-links (causal-order-eq active-equation)
		     (causal-order-eq active-equation))
	(output-links
	 (build-lists (mapcar #'make-output-eq rough-links))
	 (append output-links
		 (build-lists (mapcar #'make-output-eq rough-links))))
	(end-flag (null equation-list) (null equation-list)))
      (end-flag output-links)))


;; This function takes a list of lists and appends them in order. It returns 
;; the list.
(defun build-lists (list-input)
  (cond 
   ((null list-input) nil)
   ((null (car list-input)) (build-lists (cdr list-input)))
   ((atom (car list-input)) 
    (error "~%build-lists encountered atom in list ~%"))
   (t (append (car list-input) (build-lists (cdr list-input))))))


;; This is the dummy function for the impact signs. It takes a list as its
;; argument (the list of causal links) and just adds 'amb to every car of the
;; lists in the list.
(defun add-impact-sign (list)
  (mapcar #'(lambda (x)
	      (cons 'amb x))
	  list))


;; This function is the counterpart of mapcar only for structured lists
;; (trees). It takes the function and the structured list and returns the 
;; structured list with the function applied to all atoms in it. The third
;; argument determines the behaviour of the mapping function, ie mapcar or
;; mapc.
;; The function does not perform any error-checking.
(defun maptree (function tree map-function)
  
  ;; First define an internal closure which serves as the function argument for
  ;; a recursive mapcar-call.
  (labels
      ((back-end (element)
	 (cond 
	  
	  ;; If the element is nil return nil
	  ((null element) nil)
	  
	  ;; If the element is a list call the map-function on the list with
	  ;; the back-end function as the function argument.
	  ((listp element) (funcall map-function #'back-end element))
	  
	  ;; If the element is an atom apply the function to it.
	  ((atom element) (funcall function element)))))
    
    ;; Call the back-end function on the tree.
    (mapcar #'back-end tree)))
